import itertools
import json
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

from shillelagh.fields import Field
from shillelagh.fields import Float
from shillelagh.fields import Integer
from shillelagh.fields import Order
from shillelagh.fields import String
from shillelagh.types import Row

DELETED = range(-1, 0)


class RowIDManager:
    def __init__(self, ranges: List[range]):
        if not ranges:
            raise Exception("Argument `ranges` cannot be empty")

        self.ranges = ranges

    def __iter__(self):
        yield from itertools.chain(*self.ranges)

    def get_max_row_id(self) -> int:
        return max((r.stop - 1) for r in self.ranges)

    def check_row_id(self, row_id: int) -> None:
        for r in self.ranges:
            if r.start <= row_id < r.stop:
                raise Exception(f"Row ID {row_id} already present")

    def insert(self, row_id: Optional[int] = None) -> int:
        if row_id is None:
            max_row_id = self.get_max_row_id()
            row_id = max_row_id + 1
        else:
            self.check_row_id(row_id)

        last = self.ranges[-1]
        if last.stop == row_id:
            self.ranges[-1] = range(last.start, row_id + 1)
        else:
            self.ranges.append(range(row_id, row_id + 1))
        return row_id

    def delete(self, row_id: int) -> None:
        for i, r in enumerate(self.ranges):
            if r.start <= row_id < r.stop:
                if r.start == r.stop - 1:
                    self.ranges[i] = DELETED
                elif row_id == r.start:
                    self.ranges[i] = range(r.start + 1, r.stop)
                    self.ranges.insert(i, DELETED)
                elif row_id == r.stop - 1:
                    self.ranges[i] = range(r.start, r.stop - 1)
                    self.ranges.insert(i + 1, DELETED)
                else:
                    self.ranges[i] = range(r.start, row_id)
                    self.ranges.insert(i + 1, range(row_id + 1, r.stop))
                    self.ranges.insert(i + 1, DELETED)

                return

        raise Exception(f"Row ID {row_id} not found")


def analyse(
    data: Iterator[Row],
) -> Tuple[int, Dict[str, Order], Dict[str, Type[Field]]]:
    """Compute number of rows, order and types."""
    order: Dict[str, Order] = {}
    types: Dict[str, Type[Field]] = {}

    previous_row: Optional[Row] = None
    for i, row in enumerate(data):
        for column_name, value in row.items():
            # determine order
            if previous_row is not None:
                previous = previous_row[column_name]
                order[column_name] = update_order(
                    current_order=order.get(column_name, Order.NONE),
                    previous=previous,
                    current=value,
                    num_rows=i + 1,
                )

            # determine types
            if types.get(column_name) == String:
                continue
            elif type(value) == str:
                types[column_name] = String
            elif types.get(column_name) == Float:
                continue
            elif type(value) == float:
                types[column_name] = Float
            elif types.get(column_name) == Integer:
                continue
            else:
                types[column_name] = Integer

        previous_row = row

    num_rows = i + 1

    return num_rows, order, types


def update_order(
    current_order: Order,
    previous: Any,
    current: Any,
    num_rows: int,
) -> Order:
    if num_rows < 2 or previous is None:
        return Order.NONE

    try:
        if num_rows == 2:
            return Order.ASCENDING if current >= previous else Order.DESCENDING
        elif (
            current_order == Order.NONE
            or (current_order == Order.ASCENDING and current < previous)
            or (current_order == Order.DESCENDING and current > previous)
        ):
            return Order.NONE
    except TypeError:
        return Order.NONE

    return current_order


def quote(value: str) -> str:
    return value.replace("'", "''")


def unquote(value: str) -> str:
    return value.replace("''", "'")


def serialize(value: Any) -> str:
    return f"'{quote(json.dumps(value))}'"


def deserialize(value: str) -> Any:
    return json.loads(unquote(value[1:-1]))



import datetime

from shillelagh.fields import Blob
from shillelagh.fields import Boolean
from shillelagh.fields import Date
from shillelagh.fields import DateTime
from shillelagh.fields import Field
from shillelagh.fields import Float
from shillelagh.fields import Integer
from shillelagh.fields import Order
from shillelagh.fields import String
from shillelagh.fields import Time
from shillelagh.filters import Equal
from shillelagh.types import BINARY
from shillelagh.types import DATETIME
from shillelagh.types import NUMBER
from shillelagh.types import STRING


def test_comparison():
    field1 = Field(filters=[Equal], order=Order.ASCENDING, exact=True)
    field2 = Field(filters=[Equal], order=Order.ASCENDING, exact=True)
    field3 = Field(filters=[Equal], order=Order.ASCENDING, exact=False)

    assert field1 == field2
    assert field1 != field3
    assert field1 != 42


def test_integer():
    assert Integer.parse(1) == 1
    assert Integer.parse("1") == 1
    assert Integer.parse(None) is None


def test_blob():
    assert Blob.parse(1) == 1
    assert Blob.parse("test") == "test"
    assert Blob.parse(b"test") == b"test"
    assert Blob.parse(None) is None


def test_date():
    assert Date.parse("2020-01-01") == datetime.date(2020, 1, 1)
    assert Date.parse("2020-01-01T00:00+00:00") == datetime.date(
        2020,
        1,
        1,
    )
    assert Date.parse(None) is None
    assert Date.parse("invalid") is None


def test_time():
    assert Time.parse("12:00+00:00") == datetime.time(
        12,
        0,
        tzinfo=datetime.timezone.utc,
    )
    assert Time.parse("12:00") == datetime.time(
        12,
        0,
        tzinfo=datetime.timezone.utc,
    )
    assert Time.parse(None) is None
    assert Time.parse("invalid") is None


def test_datetime():
    assert DateTime.parse("2020-01-01T12:00+00:00") == datetime.datetime(
        2020,
        1,
        1,
        12,
        0,
        0,
        tzinfo=datetime.timezone.utc,
    )
    assert DateTime.parse(None) is None
    assert DateTime.parse("invalid") is None


def test_boolean():
    assert Boolean.parse(True) is True
    assert Boolean.parse(False) is False
    assert Boolean.parse("true") is True
    assert Boolean.parse(0) is False


def test_type_code():
    assert Integer == NUMBER
    assert Float == NUMBER
    assert String == STRING
    assert Date == DATETIME
    assert Time == DATETIME
    assert DateTime == DATETIME
    assert Blob == BINARY
    assert Boolean == NUMBER

    assert not NUMBER == 1



import json
from unittest import mock

import pytest
import requests
import requests_mock
from shillelagh.backends.apsw.dialects.gsheets import APSWGSheetsDialect
from shillelagh.backends.apsw.dialects.gsheets import extract_query
from shillelagh.exceptions import ProgrammingError
from sqlalchemy.engine import create_engine
from sqlalchemy.engine.url import make_url


def test_gsheets_dialect():
    dialect = APSWGSheetsDialect()
    assert dialect.create_connect_args(make_url("gsheets://")) == (
        (
            ":memory:",
            ["gsheetsapi"],
            {"gsheetsapi": (None, None, None, None)},
            True,
            None,
        ),
        {},
    )

    dialect = APSWGSheetsDialect(
        service_account_info={"secret": "XXX"},
        subject="user@example.com",
    )
    assert dialect.create_connect_args(make_url("gsheets://")) == (
        (
            ":memory:",
            ["gsheetsapi"],
            {"gsheetsapi": (None, None, {"secret": "XXX"}, "user@example.com")},
            True,
            None,
        ),
        {},
    )

    dialect = APSWGSheetsDialect(
        service_account_file="credentials.json",
        subject="user@example.com",
    )
    assert dialect.create_connect_args(make_url("gsheets://")) == (
        (
            ":memory:",
            ["gsheetsapi"],
            {"gsheetsapi": (None, "credentials.json", None, "user@example.com")},
            True,
            None,
        ),
        {},
    )

    mock_dbapi_connection = mock.MagicMock()
    assert dialect.get_schema_names(mock_dbapi_connection) == []


def test_get_table_names(mocker):
    get_credentials = mocker.patch(
        "shillelagh.backends.apsw.dialects.gsheets.get_credentials",
    )

    adapter = requests_mock.Adapter()
    session = requests.Session()
    session.mount("https://", adapter)
    mocker.patch(
        "shillelagh.backends.apsw.dialects.gsheets.AuthorizedSession",
        return_value=session,
    )
    adapter.register_uri(
        "GET",
        "https://www.googleapis.com/drive/v3/files?q=mimeType='application/vnd.google-apps.spreadsheet'",
        json={"files": [{"id": 1}, {"id": 2}, {"id": 3}]},
    )
    adapter.register_uri(
        "GET",
        "https://sheets.googleapis.com/v4/spreadsheets/1?includeGridData=false",
        json={
            "sheets": [
                {"properties": {"sheetId": 0}},
                {"properties": {"sheetId": 1}},
            ],
        },
    )
    adapter.register_uri(
        "GET",
        "https://sheets.googleapis.com/v4/spreadsheets/2?includeGridData=false",
        json={"sheets": [{"properties": {"sheetId": 0}}]},
    )
    adapter.register_uri(
        "GET",
        "https://sheets.googleapis.com/v4/spreadsheets/3?includeGridData=false",
        json={
            "error": {
                "code": 403,
                "message": "Google Sheets API has not been used in project 1034909279888 before or it is disabled. Enable it by visiting https://console.developers.google.com/apis/api/sheets.googleapis.com/overview?project=1034909279888 then retry. If you enabled this API recently, wait a few minutes for the action to propagate to our systems and retry.",
                "status": "PERMISSION_DENIED",
                "details": [
                    {
                        "@type": "type.googleapis.com/google.rpc.Help",
                        "links": [
                            {
                                "description": "Google developers console API activation",
                                "url": "https://console.developers.google.com/apis/api/sheets.googleapis.com/overview?project=1034909279888",
                            },
                        ],
                    },
                    {
                        "@type": "type.googleapis.com/google.rpc.ErrorInfo",
                        "reason": "SERVICE_DISABLED",
                        "domain": "googleapis.com",
                        "metadata": {
                            "consumer": "projects/1034909279888",
                            "service": "sheets.googleapis.com",
                        },
                    },
                ],
            },
        },
    )
    _logger = mocker.patch("shillelagh.backends.apsw.dialects.gsheets._logger")

    engine = create_engine("gsheets://")

    get_credentials.return_value = None
    tables = engine.table_names()
    assert tables == []

    get_credentials.return_value = "SECRET"
    tables = engine.table_names()
    assert tables == [
        "https://docs.google.com/spreadsheets/d/1/edit#gid=0",
        "https://docs.google.com/spreadsheets/d/1/edit#gid=1",
        "https://docs.google.com/spreadsheets/d/2/edit#gid=0",
    ]

    _logger.warning.assert_called_with(
        "Error loading sheets from file: %s",
        "Google Sheets API has not been used in project 1034909279888 before or it is disabled. Enable it by visiting https://console.developers.google.com/apis/api/sheets.googleapis.com/overview?project=1034909279888 then retry. If you enabled this API recently, wait a few minutes for the action to propagate to our systems and retry.",
    )


def test_drive_api_disabled(mocker):
    get_credentials = mocker.patch(
        "shillelagh.backends.apsw.dialects.gsheets.get_credentials",
    )

    adapter = requests_mock.Adapter()
    session = requests.Session()
    session.mount("https://", adapter)
    mocker.patch(
        "shillelagh.backends.apsw.dialects.gsheets.AuthorizedSession",
        return_value=session,
    )
    adapter.register_uri(
        "GET",
        "https://www.googleapis.com/drive/v3/files?q=mimeType='application/vnd.google-apps.spreadsheet'",
        json={
            "error": {
                "errors": [
                    {
                        "domain": "usageLimits",
                        "reason": "accessNotConfigured",
                        "message": "Access Not Configured. Drive API has not been used in project 1034909279888 before or it is disabled. Enable it by visiting https://console.developers.google.com/apis/api/drive.googleapis.com/overview?project=1034909279888 then retry. If you enabled this API recently, wait a few minutes for the action to propagate to our systems and retry.",
                        "extendedHelp": "https://console.developers.google.com/apis/api/drive.googleapis.com/overview?project=1034909279888",
                    },
                ],
                "code": 403,
                "message": "Access Not Configured. Drive API has not been used in project 1034909279888 before or it is disabled. Enable it by visiting https://console.developers.google.com/apis/api/drive.googleapis.com/overview?project=1034909279888 then retry. If you enabled this API recently, wait a few minutes for the action to propagate to our systems and retry.",
            },
        },
    )

    engine = create_engine("gsheets://")

    get_credentials.return_value = "SECRET"
    with pytest.raises(ProgrammingError) as excinfo:
        engine.table_names()

    assert (
        str(excinfo.value)
        == "Access Not Configured. Drive API has not been used in project 1034909279888 before or it is disabled. Enable it by visiting https://console.developers.google.com/apis/api/drive.googleapis.com/overview?project=1034909279888 then retry. If you enabled this API recently, wait a few minutes for the action to propagate to our systems and retry."
    )


def test_extract_query():
    assert extract_query(make_url("gsheets://")) == {}
    assert extract_query(make_url("gsheets://host")) == {}
    assert extract_query(make_url("gsheets://?foo=bar")) == {"foo": "bar"}
    assert extract_query(make_url("gsheets:///?foo=bar")) == {"foo": "bar"}



from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

from shillelagh.backends.apsw.dialects.base import APSWDialect
from sqlalchemy.engine.url import URL


class APSWSafeDialect(APSWDialect):
    def __init__(
        self,
        adapters: Optional[List[str]] = None,
        adapter_args: Optional[Dict[str, Any]] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._adapters = adapters
        self._adapter_args = adapter_args
        self._safe = True

    def create_connect_args(
        self,
        url: URL,
    ) -> Tuple[
        Tuple[str, Optional[List[str]], Optional[Dict[str, Any]], bool, Optional[str]],
        Dict[str, Any],
    ]:
        return (
            (
                ":memory:",
                self._adapters,
                self._adapter_args,
                True,
                self.isolation_level,
            ),
            {},
        )



import logging
import urllib.parse
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

from google.auth.credentials import Credentials
from google.auth.transport.requests import AuthorizedSession
from shillelagh.adapters.api.gsheets import get_credentials
from shillelagh.backends.apsw.dialects.base import APSWDialect
from shillelagh.exceptions import ProgrammingError
from sqlalchemy.engine.url import URL
from sqlalchemy.pool.base import _ConnectionFairy

_logger = logging.getLogger(__name__)


def extract_query(url: URL) -> Dict[str, str]:
    """
    Extract the query from the SQLAlchemy URL.

    There's a bug in how SQLAlchemy handles URLs without hosts:

        >>> from sqlalchemy.engine.url import make_url
        >>> url = make_url("gsheets://")
        >>> url.query["subject"] = "user@example.com"
        >>> url
        gsheets://?subject=user%40example.com
        >>> make_url(str(url)).query
        {}
        >>> make_url(str(url)).host
        '?subject=user%40example.com'

    """
    if url.query:
        return dict(url.query)
    if url.host and url.host.startswith("?"):
        return dict(urllib.parse.parse_qsl(url.host[1:]))
    return {}


class APSWGSheetsDialect(APSWDialect):
    """Drop-in replacement for gsheetsdb."""

    name = "gsheets"

    def __init__(
        self,
        access_token: Optional[str] = None,
        service_account_file: Optional[str] = None,
        service_account_info: Optional[Dict[str, Any]] = None,
        subject: Optional[str] = None,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)

        self.access_token = access_token
        self.service_account_file = service_account_file
        self.service_account_info = service_account_info
        self.subject = subject

    def create_connect_args(
        self,
        url: URL,
    ) -> Tuple[
        Tuple[str, Optional[List[str]], Optional[Dict[str, Any]], bool, Optional[str]],
        Dict[str, Any],
    ]:
        query = extract_query(url)
        adapter_args: Dict[str, Any] = {
            "gsheetsapi": (
                query.get("access_token", self.access_token),
                query.get("service_account_file", self.service_account_file),
                query.get("service_account_info", self.service_account_info),
                query.get("subject", self.subject),
            ),
        }

        return (
            ":memory:",
            ["gsheetsapi"],
            adapter_args,
            True,
            self.isolation_level,
        ), {}

    def get_schema_names(
        self, connection: _ConnectionFairy, **kwargs: Any
    ) -> List[str]:
        return []

    def get_table_names(
        self, connection: _ConnectionFairy, schema: str = None, **kwargs: Any
    ) -> List[str]:
        query = extract_query(connection.url) if hasattr(connection, "url") else {}
        credentials = get_credentials(
            query.get("access_token", self.access_token),
            query.get("service_account_file", self.service_account_file),
            self.service_account_info,
            query.get("subject", self.subject),
        )
        if not credentials:
            return []

        session = AuthorizedSession(credentials)

        tables = []
        response = session.get(
            "https://www.googleapis.com/drive/v3/files?q=mimeType='application/vnd.google-apps.spreadsheet'",
        )
        payload = response.json()
        if "error" in payload:
            raise ProgrammingError(payload["error"]["message"])

        files = payload["files"]
        for file in files:
            spreadsheet_id = file["id"]
            response = session.get(
                f"https://sheets.googleapis.com/v4/spreadsheets/{spreadsheet_id}?includeGridData=false",
            )
            payload = response.json()
            if "error" in payload:
                _logger.warning(
                    "Error loading sheets from file: %s",
                    payload["error"]["message"],
                )
                continue

            sheets = payload["sheets"]
            for sheet in sheets:
                sheet_id = sheet["properties"]["sheetId"]
                tables.append(
                    f"https://docs.google.com/spreadsheets/d/{spreadsheet_id}/edit#gid={sheet_id}",
                )

        return tables



from datetime import date
from datetime import datetime
from datetime import time
from datetime import timezone

from shillelagh.backends.apsw.db import connect
from shillelagh.types import Binary
from shillelagh.types import Date
from shillelagh.types import DateFromTicks
from shillelagh.types import Time
from shillelagh.types import TimeFromTicks
from shillelagh.types import Timestamp
from shillelagh.types import TimestampFromTicks


def test_types():
    connection = connect(":memory:")
    cursor = connection.cursor()

    cursor.execute(
        """
        CREATE TABLE test_types (
            type_date DATE,
            type_time TIME,
            type_timestamp TIMESTAMP,
            type_binary BLOB
        )
    """,
    )
    cursor.execute(
        "INSERT INTO test_types VALUES (?, ?, ?, ?)",
        (
            Date(2020, 1, 1),
            Time(0, 0, 0),
            Timestamp(2020, 1, 1, 0, 0, 0),
            Binary("🦥"),
        ),
    )
    cursor.execute("SELECT * FROM test_types")
    row = cursor.fetchone()
    assert row == (
        date(2020, 1, 1),
        time(0, 0, tzinfo=timezone.utc),
        datetime(2020, 1, 1, 0, 0, tzinfo=timezone.utc),
        b"\xf0\x9f\xa6\xa5",
    )

    cursor.execute(
        "INSERT INTO test_types VALUES (?, ?, ?, ?)",
        (
            DateFromTicks(1),
            TimeFromTicks(2),
            TimestampFromTicks(3),
            Binary("🦥"),
        ),
    )
    cursor.execute("SELECT * FROM test_types")
    rows = cursor.fetchall()
    assert rows == [
        (
            date(2020, 1, 1),
            time(0, 0, tzinfo=timezone.utc),
            datetime(2020, 1, 1, 0, 0, tzinfo=timezone.utc),
            b"\xf0\x9f\xa6\xa5",
        ),
        (
            date(1970, 1, 1),
            time(0, 0, 2, tzinfo=timezone.utc),
            datetime(1970, 1, 1, 0, 0, 3, tzinfo=timezone.utc),
            b"\xf0\x9f\xa6\xa5",
        ),
    ]



from shillelagh.backends.apsw.dialects.safe import APSWSafeDialect
from sqlalchemy.engine.url import make_url


def test_safe_dialect(fs):
    dialect = APSWSafeDialect()
    assert dialect.create_connect_args(make_url("shillelagh+safe://")) == (
        (
            ":memory:",
            None,
            None,
            True,
            None,
        ),
        {},
    )



from datetime import datetime
from datetime import timezone

import pytest
from shillelagh.adapters.api.socrata import build_sql
from shillelagh.adapters.api.socrata import convert_rows
from shillelagh.adapters.api.socrata import quote
from shillelagh.backends.apsw.db import connect
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import DateTime
from shillelagh.fields import String
from shillelagh.filters import Range

from ...fakes import cdc_data_response
from ...fakes import cdc_metadata_response


def test_socrata(requests_mock):
    metadata_url = "https://data.cdc.gov/api/views/unsk-b7fc"
    requests_mock.get(metadata_url, json=cdc_metadata_response)

    data_url = (
        "https://data.cdc.gov/resource/unsk-b7fc.json?"
        "%24query=SELECT+%2A+WHERE+location+%3D+%27US%27+ORDER+BY+date+DESC"
    )
    requests_mock.get(data_url, json=cdc_data_response)

    connection = connect(":memory:")
    cursor = connection.cursor()
    sql = """
        SELECT date, administered_dose1_recip_4
        FROM "https://data.cdc.gov/resource/unsk-b7fc.json"
        WHERE location = 'US'
        ORDER BY date DESC
        LIMIT 7
    """
    data = list(cursor.execute(sql))
    assert data == [
        (datetime(2021, 6, 3, 0, 0, tzinfo=timezone.utc), 63.0),
        (datetime(2021, 6, 2, 0, 0, tzinfo=timezone.utc), 62.9),
        (datetime(2021, 6, 1, 0, 0, tzinfo=timezone.utc), 62.8),
        (datetime(2021, 5, 31, 0, 0, tzinfo=timezone.utc), 62.7),
        (datetime(2021, 5, 30, 0, 0, tzinfo=timezone.utc), 62.6),
        (datetime(2021, 5, 29, 0, 0, tzinfo=timezone.utc), 62.4),
        (datetime(2021, 5, 28, 0, 0, tzinfo=timezone.utc), 62.2),
    ]


def test_socrata_no_data(requests_mock):
    metadata_url = "https://data.cdc.gov/api/views/unsk-b7fc"
    requests_mock.get(metadata_url, json=cdc_metadata_response)

    data_url = (
        "https://data.cdc.gov/resource/unsk-b7fc.json?"
        "%24query=SELECT+%2A+WHERE+location+%3D+%27BR%27+ORDER+BY+date+DESC"
    )
    requests_mock.get(data_url, json=[])

    connection = connect(":memory:")
    cursor = connection.cursor()
    sql = """
        SELECT date, administered_dose1_recip_4
        FROM "https://data.cdc.gov/resource/unsk-b7fc.json"
        WHERE location = 'BR'
        ORDER BY date DESC
        LIMIT 7
    """
    data = list(cursor.execute(sql))
    assert data == []


def test_build_sql():
    sql = build_sql({"a": Range(1, 10, False, True)}, [])
    assert sql == "SELECT * WHERE a > 1 AND a <= 10"

    sql = build_sql({"a": Range(1, None, True, False)}, [])
    assert sql == "SELECT * WHERE a >= 1"

    sql = build_sql({"a": Range(None, 10, True, False)}, [])
    assert sql == "SELECT * WHERE a < 10"

    sql = build_sql({}, [])
    assert sql == "SELECT *"

    with pytest.raises(ProgrammingError) as excinfo:
        build_sql({"a": [1, 2, 3]}, [])
    assert str(excinfo.value) == "Invalid filter: [1, 2, 3]"


def test_socrata_impossible(requests_mock):
    metadata_url = "https://data.cdc.gov/api/views/unsk-b7fc"
    requests_mock.get(metadata_url, json=cdc_metadata_response)

    connection = connect(":memory:")
    cursor = connection.cursor()
    sql = """
        SELECT date, administered_dose1_recip_4
        FROM "https://data.cdc.gov/resource/unsk-b7fc.json"
        WHERE location = 'US' AND location = 'AZ'
        ORDER BY date DESC
        LIMIT 7
    """
    data = list(cursor.execute(sql))
    assert data == []


def test_quote():
    assert quote(1) == "1"
    assert quote(1.0) == "1.0"
    assert quote("one") == "'one'"
    assert (
        quote(datetime(2021, 6, 3, 7, 0, tzinfo=timezone.utc))
        == "'2021-06-03T07:00:00+00:00'"
    )

    with pytest.raises(Exception) as excinfo:
        quote([1])
    assert str(excinfo.value) == "Can't quote value: [1]"


def test_convert_rows():
    columns = {"a": String(), "b": DateTime()}
    rows = [
        {"a": "one", "b": "2021-06-03T07:00:00+00:00"},
    ]
    converted = convert_rows(columns, rows)
    assert list(converted) == [
        {"a": "one", "b": datetime(2021, 6, 3, 7, 0, tzinfo=timezone.utc)},
    ]



import re
import urllib.parse
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

import requests_cache
from shillelagh.adapters.base import Adapter
from shillelagh.exceptions import ImpossibleFilterError
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import DateTime
from shillelagh.fields import Field
from shillelagh.fields import Float
from shillelagh.fields import Order
from shillelagh.fields import String
from shillelagh.filters import Equal
from shillelagh.filters import Filter
from shillelagh.filters import Impossible
from shillelagh.filters import Range
from shillelagh.types import RequestedOrder
from shillelagh.types import Row
from typing_extensions import TypedDict


path_regex = re.compile(r"/resource/\w{4}-\w{4}.json")


class MetadataColumn(TypedDict):
    id: int
    name: str
    dataTypeName: str
    description: str
    fieldName: str
    position: int
    renderTypeName: str
    tableColumnId: int
    cachedContents: Dict[str, Any]
    format: Dict[str, Any]


type_map: Dict[str, Tuple[Type[Field], List[Type[Filter]]]] = {
    "calendar_date": (DateTime, [Range]),
    "number": (Float, [Range]),
    "text": (String, [Equal]),
}


def get_field(col: MetadataColumn) -> Field:
    class_, filters = type_map.get(col["dataTypeName"], (String, [Equal]))
    return class_(
        filters=filters,
        order=Order.ANY,
        exact=True,
    )


def quote(value: Any) -> str:
    if isinstance(value, (int, float)):
        return str(value)
    if isinstance(value, str):
        escaped_value = value.replace("'", "''")
        return f"'{escaped_value}'"
    if hasattr(value, "isoformat"):
        return f"'{value.isoformat()}'"

    raise Exception(f"Can't quote value: {value}")


def build_sql(
    bounds: Dict[str, Filter],
    order: List[Tuple[str, RequestedOrder]],
) -> str:
    sql = "SELECT *"

    conditions = []
    for column_name, filter_ in bounds.items():
        if isinstance(filter_, Impossible):
            raise ImpossibleFilterError()
        if isinstance(filter_, Equal):
            conditions.append(f"{column_name} = {quote(filter_.value)}")
        elif isinstance(filter_, Range):
            if filter_.start is not None:
                op = ">=" if filter_.include_start else ">"
                conditions.append(f"{column_name} {op} {quote(filter_.start)}")
            if filter_.end is not None:
                op = "<=" if filter_.include_end else "<"
                conditions.append(f"{column_name} {op} {quote(filter_.end)}")
        else:
            raise ProgrammingError(f"Invalid filter: {filter_}")
    if conditions:
        sql = f"{sql} WHERE {' AND '.join(conditions)}"

    column_order: List[str] = []
    for column_name, requested_order in order:
        desc = " DESC" if requested_order == Order.DESCENDING else ""
        column_order.append(f"{column_name}{desc}")
    if column_order:
        sql = f"{sql} ORDER BY {', '.join(column_order)}"

    return sql


def convert_rows(columns: Dict[str, Field], rows: List[Row]) -> Iterator[Row]:
    for row in rows:
        yield {
            column_name: columns[column_name].parse(value)
            for column_name, value in row.items()
        }


class SocrataAPI(Adapter):

    safe = True

    @staticmethod
    def supports(uri: str) -> bool:
        """https://data.cdc.gov/resource/unsk-b7fc.json"""
        parsed = urllib.parse.urlparse(uri)
        return bool(path_regex.match(parsed.path))

    @staticmethod
    def parse_uri(uri: str) -> Tuple[str, str, Optional[str]]:
        parsed = urllib.parse.urlparse(uri)
        dataset_id = Path(parsed.path).stem
        query_string = urllib.parse.parse_qs(parsed.query)
        app_token = (
            query_string["$$app_token"][0] if "$$app_token" in query_string else None
        )

        return (
            parsed.netloc,
            dataset_id,
            app_token,
        )

    def __init__(self, netloc: str, dataset_id: str, app_token: Optional[str]):
        self.netloc = netloc
        self.dataset_id = dataset_id
        self.app_token = app_token

        self._session = requests_cache.CachedSession(
            cache_name="socrata_cache",
            backend="sqlite",
            expire_after=180,
        )

        self._set_columns()

    def _set_columns(self) -> None:
        url = f"https://{self.netloc}/api/views/{self.dataset_id}"
        response = self._session.get(url)
        payload = response.json()
        self.columns = {col["fieldName"]: get_field(col) for col in payload["columns"]}

    def get_columns(self) -> Dict[str, Field]:
        return self.columns

    def get_data(
        self,
        bounds: Dict[str, Filter],
        order: List[Tuple[str, RequestedOrder]],
    ) -> Iterator[Row]:
        try:
            sql = build_sql(bounds, order)
        except ImpossibleFilterError:
            return

        url = f"https://{self.netloc}/resource/{self.dataset_id}.json"
        headers = {"X-App-Token": self.app_token} if self.app_token else {}
        response = self._session.get(url, params={"$query": sql}, headers=headers)
        results = response.json()
        rows = convert_rows(self.columns, results)

        for i, row in enumerate(rows):
            row["rowid"] = i
            yield row



import datetime
import json
import urllib.parse
from typing import Any
from typing import Callable
from typing import cast
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type

import google.oauth2.credentials
import google.oauth2.service_account
from google.auth.credentials import Credentials
from google.auth.transport.requests import AuthorizedSession
from requests import Session
from shillelagh.adapters.base import Adapter
from shillelagh.exceptions import ProgrammingError
from shillelagh.fields import Boolean
from shillelagh.fields import Date
from shillelagh.fields import DateTime
from shillelagh.fields import Field
from shillelagh.fields import Float
from shillelagh.fields import Order
from shillelagh.fields import String
from shillelagh.fields import Time
from shillelagh.filters import Equal
from shillelagh.filters import Filter
from shillelagh.filters import Impossible
from shillelagh.filters import Range
from shillelagh.types import RequestedOrder
from shillelagh.types import Row
from typing_extensions import Literal
from typing_extensions import TypedDict

# Google API scopes for authentication
SCOPES = [
    "https://www.googleapis.com/auth/drive.readonly",
]

JSON_PAYLOAD_PREFIX = ")]}'\n"


class UrlArgs(TypedDict, total=False):
    headers: int
    gid: int
    sheet: str


class QueryResultsColumn(TypedDict, total=False):
    id: str
    label: str
    type: str
    pattern: str  # optional


class QueryResultsCell(TypedDict, total=False):
    v: Any
    f: str  # optional


class QueryResultsRow(TypedDict):
    c: List[QueryResultsCell]


class QueryResultsTable(TypedDict):
    cols: List[QueryResultsColumn]
    rows: List[QueryResultsRow]
    parsedNumHeaders: int


class QueryResultsError(TypedDict):
    reason: str
    message: str
    detailed_message: str


class QueryResults(TypedDict, total=False):
    """
    Query results from the Google API.

    Successful query:

    {
        "version": "0.6",
        "reqId": "0",
        "status": "ok",
        "sig": "1453301915",
        "table": {
            "cols": [
                {"id": "A", "label": "country", "type": "string"},
                {"id": "B", "label": "cnt", "type": "number", "pattern": "General"},
            ],
            "rows": [{"c": [{"v": "BR"}, {"v": 1.0, "f": "1"}]}],
            "parsedNumHeaders": 0,
        },
    }

    Failed:

    {
        "version": "0.6",
        "reqId": "0",
        "status": "error",
        "errors": [
            {
                "reason": "invalid_query",
                "message": "INVALID_QUERY",
                "detailed_message": "Invalid query: NO_COLUMN: C",
            }
        ],
    }
    """

    version: str
    reqId: str
    status: Literal["ok", "error"]
    sig: str
    table: QueryResultsTable
    errors: List[QueryResultsError]


type_map: Dict[str, Tuple[Type[Field], List[Type[Filter]]]] = {
    "string": (String, [Equal]),
    "number": (Float, [Range]),
    "boolean": (Boolean, [Equal]),
    "date": (Date, [Range]),
    "datetime": (DateTime, [Range]),
    "timeofday": (Time, [Range]),
}


def get_field(col: QueryResultsColumn) -> Field:
    class_, filters = type_map.get(col["type"], (String, [Equal]))
    return class_(
        filters=filters,
        order=Order.ANY,
        exact=True,
    )


def quote(value: Any) -> str:
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, (int, float)):
        return str(value)
    if isinstance(value, str):
        escaped_value = value.replace("'", "''")
        return f"'{escaped_value}'"
    if isinstance(value, datetime.datetime):
        return f"datetime '{value}'"
    if isinstance(value, datetime.date):
        return f"date '{value}'"
    if isinstance(value, datetime.time):
        return f"timeofday '{value}'"

    raise Exception(f"Can't quote value: {value}")


def format_error_message(errors: List[QueryResultsError]) -> str:
    return "\n\n".join(error["detailed_message"] for error in errors)


def get_url(
    uri: str,
    headers: int = 0,
    gid: int = 0,
    sheet: Optional[str] = None,
) -> str:
    """Return API URL given the spreadsheet URL."""
    parts = urllib.parse.urlparse(uri)

    # strip /edit
    path = parts.path[: -len("/edit")] if parts.path.endswith("/edit") else parts.path

    # add the gviz endpoint
    path = "/".join((path.rstrip("/"), "gviz/tq"))

    qs = urllib.parse.parse_qs(parts.query)
    if "headers" in qs:
        headers = int(qs["headers"][-1])
    if "gid" in qs:
        gid = int(qs["gid"][-1])
    if "sheet" in qs:
        sheet = qs["sheet"][-1]

    if parts.fragment.startswith("gid="):
        gid = int(parts.fragment[len("gid=") :])

    args: UrlArgs = {}
    if headers > 0:
        args["headers"] = headers
    if sheet is not None:
        args["sheet"] = sheet
    else:
        args["gid"] = gid
    params = urllib.parse.urlencode(args)

    return urllib.parse.urlunparse(
        (parts.scheme, parts.netloc, path, None, params, None),
    )


def parse_datetime(value: str) -> datetime.datetime:
    """Parse a string like 'Date(2018,0,1,0,0,0)'."""
    args = [int(number) for number in value[len("Date(") : -1].split(",")]
    args[1] += 1  # month is zero indexed in the response
    return datetime.datetime(*args, tzinfo=datetime.timezone.utc)  # type: ignore


def parse_date(value: str) -> datetime.date:
    """Parse a string like 'Date(2018,0,1)'."""
    args = [int(number) for number in value[len("Date(") : -1].split(",")]
    args[1] += 1  # month is zero indexed in the response
    return datetime.date(*args)


def parse_timeofday(values: List[int]) -> datetime.time:
    """Parse time of day as returned from the API."""
    return datetime.time(*values, tzinfo=datetime.timezone.utc)  # type: ignore


converters: Dict[str, Callable[[Any], Any]] = {
    "string": lambda v: v,
    "number": lambda v: v,
    "boolean": bool,
    "date": parse_date,
    "datetime": parse_datetime,
    "timeofday": parse_timeofday,
}


def convert_rows(
    cols: List[QueryResultsColumn],
    rows: List[QueryResultsRow],
) -> Iterator[List[Any]]:
    """Convert custom Google sheets types."""
    row_converters: Optional[List[Callable[[Any], Any]]] = None
    for row in rows:
        if row_converters is None:
            row_converters = [converters[col["type"]] for col in cols]
        yield [
            converter(cell["v"]) if cell else None
            for cell, converter in zip(row["c"], row_converters)
        ]


def get_credentials(
    access_token: Optional[str],
    service_account_file: Optional[str],
    service_account_info: Optional[Dict[str, Any]],
    subject: Optional[str],
) -> Optional[Credentials]:
    if access_token:
        return google.oauth2.credentials.Credentials(access_token)

    if service_account_file:
        return google.oauth2.service_account.Credentials.from_service_account_file(
            service_account_file,
            scopes=SCOPES,
            subject=subject,
        )

    if service_account_info:
        return google.oauth2.service_account.Credentials.from_service_account_info(
            service_account_info,
            scopes=SCOPES,
            subject=subject,
        )

    return None


class GSheetsAPI(Adapter):

    safe = True

    @staticmethod
    def supports(uri: str) -> bool:
        parsed = urllib.parse.urlparse(uri)
        return parsed.netloc == "docs.google.com" and parsed.path.startswith(
            "/spreadsheets/",
        )

    @staticmethod
    def parse_uri(uri: str) -> Tuple[str]:
        return (uri,)

    def __init__(
        self,
        uri: str,
        access_token: Optional[str] = None,
        service_account_file: Optional[str] = None,
        service_account_info: Optional[Dict[str, Any]] = None,
        subject: Optional[str] = None,
    ):
        self.url = get_url(uri)
        self.credentials = get_credentials(
            access_token,
            service_account_file,
            service_account_info,
            subject,
        )

        self._offset = 0
        self._set_columns()

    def _get_session(self) -> Session:
        return cast(
            Session,
            AuthorizedSession(self.credentials) if self.credentials else Session(),
        )

    def _run_query(self, sql: str) -> QueryResults:
        quoted_sql = urllib.parse.quote(sql, safe="/()")
        url = f"{self.url}&tq={quoted_sql}"
        headers = {"X-DataSource-Auth": "true"}

        session = self._get_session()
        response = session.get(url, headers=headers)
        if response.encoding is None:
            response.encoding = "utf-8"

        if response.status_code != 200:
            raise ProgrammingError(response.text)

        if response.text.startswith(JSON_PAYLOAD_PREFIX):
            result = json.loads(response.text[len(JSON_PAYLOAD_PREFIX) :])
        else:
            try:
                result = response.json()
            except json.decoder.JSONDecodeError:
                raise ProgrammingError(
                    "Response from Google is not valid JSON. Please verify that you "
                    "have the proper credentials to access the spreadsheet.",
                )

        if result["status"] == "error":
            raise ProgrammingError(format_error_message(result["errors"]))

        return cast(QueryResults, result)

    def _set_columns(self) -> None:
        results = self._run_query("SELECT * LIMIT 1")
        cols = results["table"]["cols"]
        rows = results["table"]["rows"]

        # if the columns have no labels, use the first row as the labels if
        # it exists; otherwise use just the column letters
        if all(col["label"] == "" for col in cols):
            if rows:
                self._offset = 1
                for col, cell in zip(cols, rows[0]["c"]):
                    col["label"] = cell["v"]
            else:
                for col in cols:
                    col["label"] = col["id"]

        self._column_map = {col["label"].strip(): col["id"] for col in cols}
        self.columns = {
            col["label"].strip(): get_field(col) for col in cols if col["label"].strip()
        }

    def get_columns(self) -> Dict[str, Field]:
        return self.columns

    def _build_sql(
        self,
        bounds: Dict[str, Filter],
        order: List[Tuple[str, RequestedOrder]],
    ) -> str:
        sql = "SELECT *"

        conditions = []
        for column_name, filter_ in bounds.items():
            id_ = self._column_map[column_name]
            if isinstance(filter_, Impossible):
                conditions.append("1 = 0")
            elif isinstance(filter_, Equal):
                conditions.append(f"{id_} = {quote(filter_.value)}")
            elif isinstance(filter_, Range):
                if filter_.start is not None:
                    op = ">=" if filter_.include_start else ">"
                    conditions.append(f"{id_} {op} {quote(filter_.start)}")
                if filter_.end is not None:
                    op = "<=" if filter_.include_end else "<"
                    conditions.append(f"{id_} {op} {quote(filter_.end)}")
            else:
                raise ProgrammingError(f"Invalid filter: {filter_}")
        if conditions:
            sql = f"{sql} WHERE {' AND '.join(conditions)}"

        column_order: List[str] = []
        for column_name, requested_order in order:
            desc = " DESC" if requested_order == Order.DESCENDING else ""
            column_order.append(f"{column_name}{desc}")
        if column_order:
            sql = f"{sql} ORDER BY {', '.join(column_order)}"
        if self._offset > 0:
            sql = f"{sql} OFFSET {self._offset}"

        return sql

    def get_data(
        self,
        bounds: Dict[str, Filter],
        order: List[Tuple[str, RequestedOrder]],
    ) -> Iterator[Row]:
        sql = self._build_sql(bounds, order)
        results = self._run_query(sql)
        cols = results["table"]["cols"]
        rows = convert_rows(cols, results["table"]["rows"])

        for i, row in enumerate(rows):
            data = dict(zip(self.columns, row))
            data["rowid"] = i
            yield data



from unittest import mock

import pytest
from shillelagh.backends.apsw.dialects.base import APSWDialect
from shillelagh.exceptions import ProgrammingError
from sqlalchemy import create_engine
from sqlalchemy import func
from sqlalchemy import MetaData
from sqlalchemy import select
from sqlalchemy import Table

from ....fakes import FakeAdapter
from ....fakes import FakeEntryPoint


def test_create_engine(mocker):
    entry_points = [FakeEntryPoint("dummy", FakeAdapter)]
    mocker.patch(
        "shillelagh.backends.apsw.db.iter_entry_points",
        return_value=entry_points,
    )

    engine = create_engine("shillelagh://")

    table = Table("dummy://", MetaData(bind=engine), autoload=True)
    query = select([func.sum(table.columns.pets)], from_obj=table)
    assert query.scalar() == 3


def test_create_engine_no_adapters(mocker):
    engine = create_engine("shillelagh://")

    with pytest.raises(ProgrammingError) as excinfo:
        Table("dummy://", MetaData(bind=engine), autoload=True)
    assert str(excinfo.value) == "Unsupported table: dummy://"


def test_dialect_ping():
    mock_dbapi_connection = mock.MagicMock()
    dialect = APSWDialect()
    assert dialect.do_ping(mock_dbapi_connection) is True
